"use strict";

Object.defineProperty(exports, "__esModule", {
    value: true
});

var _asyncToGenerator2 = require("babel-runtime/helpers/asyncToGenerator");

var _asyncToGenerator3 = _interopRequireDefault(_asyncToGenerator2);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/*
**  GraphQL-Tools-Subscribe -- Subscription Framework for GraphQL-Tools
**  Copyright (c) 2016-2018 Ralf S. Engelschall <rse@engelschall.com>
**
**  Permission is hereby granted, free of charge, to any person obtaining
**  a copy of this software and associated documentation files (the
**  "Software"), to deal in the Software without restriction, including
**  without limitation the rights to use, copy, modify, merge, publish,
**  distribute, sublicense, and/or sell copies of the Software, and to
**  permit persons to whom the Software is furnished to do so, subject to
**  the following conditions:
**
**  The above copyright notice and this permission notice shall be included
**  in all copies or substantial portions of the Software.
**
**  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
**  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
**  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
**  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
**  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
**  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
**  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

/*  the API mixin class  */
class gtsSubscription {
    /*  the factory methods for resolving "Subscription" root type  */
    schemaSubscription() {
        return "" + "#   Access the GraphQL Subscription service.\n" + "_Subscription: _Subscription";
    }
    resolverSubscription() {
        return (() => {
            var _ref = (0, _asyncToGenerator3.default)(function* (obj, args, ctx, info) {
                /*  sanity check usage  */
                if (!ctx.scope) throw new Error("_Subscription: only allowed on WebSocket connection");

                /*  provide a wrapper object containing the unique connection id  */
                return {
                    cid: ctx.scope.connection !== null ? ctx.scope.connection.cid : "<none>"
                };
            });

            return function (_x, _x2, _x3, _x4) {
                return _ref.apply(this, arguments);
            };
        })();
    }

    /*  the factory methods for resolving "subscribe" (mutation)  */
    schemaSubscribe() {
        return "" + "#   Subscribe to the current GraphQL query\n" + "#   and retrieve its unique subscription id (SID).\n" + "subscribe: UUID!";
    }
    resolverSubscribe() {
        var _this = this;

        return (() => {
            var _ref2 = (0, _asyncToGenerator3.default)(function* (obj, args, ctx, info) {
                /*  sanity check usage  */
                if (!ctx.scope) throw new Error("subscribe: only allowed on WebSocket connection");
                if (info && info.operation && info.operation.operation !== "query") throw new Error("subscribe: only allowed in GraphQL \"query\" operation");

                /*  determine parameters  */
                let cid = obj.cid;
                let sid = ctx.scope.sid;

                /*  add to record  */
                yield _this.keyval.acquire();
                yield _this.keyval.put(`sid:${sid},cid:${cid}`, "subscribed");
                yield _this.keyval.release();

                /*  drop an event about the subscription  */
                ctx.scope.api.emit("scope-subscribe", { sid: sid, cid: cid });
                ctx.scope.api.emit("debug", `scope-subscribe sid=${sid} cid=${cid}`);

                /*  return result  */
                return sid;
            });

            return function (_x5, _x6, _x7, _x8) {
                return _ref2.apply(this, arguments);
            };
        })();
    }

    /*  the factory methods for resolving "unsubscribe" (mutation)  */
    schemaUnsubscribe(prefix = "") {
        return "" + "#   Unsubscribe from the GraphQL query\n" + "#   identified by the unique subscription id (SID).\n" + "unsubscribe(sid: UUID!): Void\n";
    }
    resolverUnsubscribe() {
        var _this2 = this;

        return (() => {
            var _ref3 = (0, _asyncToGenerator3.default)(function* (obj, args, ctx, info) {
                /*  sanity check usage  */
                if (!ctx.scope) throw new Error("unsubscribe: only allowed on WebSocket connection");
                if (info && info.operation && info.operation.operation !== "mutation") throw new Error("unsubscribe: only allowed in GraphQL \"mutation\" operation");

                /*  determine parameters  */
                let cid = obj.cid;
                let sid = args.sid;

                /*  delete from record  */
                yield _this2.keyval.acquire();
                let subscription = yield _this2.keyval.get(`sid:${sid},cid:${cid}`);
                if (subscription === undefined) {
                    yield _this2.keyval.release();
                    throw new Error(`no such subscription "${sid}"`);
                }
                yield _this2.keyval.del(`cid:${cid},sid:${sid}`);
                yield _this2.keyval.release();

                /*  drop an event about the unsubscription  */
                ctx.scope.api.emit("scope-unsubscribe", { sid: sid, cid: cid });
                ctx.scope.api.emit("debug", `scope-unsubscribe sid=${sid} cid=${cid}`);

                /*  return no result  */
                return {};
            });

            return function (_x9, _x10, _x11, _x12) {
                return _ref3.apply(this, arguments);
            };
        })();
    }

    /*  the factory methods for resolving "subscriptions" (query)  */
    schemaSubscriptions(prefix = "") {
        return "" + "#   Retrieve all (or just the outdated) active subscription ids (SIDs).\n" + "subscriptions: [UUID]!\n";
    }
    resolverSubscriptions() {
        var _this3 = this;

        return (() => {
            var _ref4 = (0, _asyncToGenerator3.default)(function* (obj, args, ctx, info) {
                /*  sanity check usage  */
                if (!ctx.scope) throw new Error("subscription: only allowed on WebSocket connection");

                /*  determine parameters  */
                let cid = obj.cid;

                /*  determine subscriptions  */
                yield _this3.keyval.acquire();
                let keys = yield _this3.keyval.keys(`sid:*,cid:${cid}`);
                let sids = keys.map(function (key) {
                    return key.replace(/^sid:(.+?),cid:.+$/, "$1");
                });
                yield _this3.keyval.release();

                /*  drop an event about the subscriptions  */
                ctx.scope.api.emit("scope-subscriptions", { cid: cid, sids: sids });
                ctx.scope.api.emit("debug", `scope-subscriptions sids=${sids.join(",")} cid=${cid}`);

                /*  return result  */
                return sids;
            });

            return function (_x13, _x14, _x15, _x16) {
                return _ref4.apply(this, arguments);
            };
        })();
    }

    /*  the factory methods for resolving "pause" (mutation)  */
    schemaPause(prefix = "") {
        return "" + "#   Pause active subscription,\n" + "#   identified by the unique subscription id (SID).\n" + "pause(sid: UUID!): Void\n";
    }
    resolverPause() {
        var _this4 = this;

        return (() => {
            var _ref5 = (0, _asyncToGenerator3.default)(function* (obj, args, ctx, info) {
                /*  sanity check usage  */
                if (!ctx.scope) throw new Error("pause: only allowed on WebSocket connection");
                if (info && info.operation && info.operation.operation !== "mutation") throw new Error("pause: only allowed in GraphQL \"mutation\" operation");

                /*  determine parameters  */
                let cid = obj.cid;
                let sid = args.sid;

                /*  determine subscriptions  */
                yield _this4.keyval.acquire();
                let subscription = yield _this4.keyval.get(`sid:${sid},cid:${cid}`);
                if (subscription === undefined) {
                    yield _this4.keyval.release();
                    throw new Error(`no such subscription "${sid}"`);
                } else if (subscription === "paused") {
                    yield _this4.keyval.release();
                    throw new Error(`subscription "${sid}" already paused`);
                }
                yield _this4.keyval.put(`sid:${sid},cid:${cid}`, "paused");
                yield _this4.keyval.release();

                /*  drop an event about the pausing  */
                ctx.scope.api.emit("scope-pause", { sid: sid, cid: cid });
                ctx.scope.api.emit("debug", `scope-pause sid=${sid} cid=${cid}`);

                /*  return no result  */
                return {};
            });

            return function (_x17, _x18, _x19, _x20) {
                return _ref5.apply(this, arguments);
            };
        })();
    }

    /*  the factory methods for resolving "resume" (mutation)  */
    schemaResume(prefix = "") {
        return "" + "#   Resume active subscription,\n" + "#   identified by the unique subscription id (SID).\n" + "resume(sid: UUID!): Void\n";
    }
    resolverResume() {
        var _this5 = this;

        return (() => {
            var _ref6 = (0, _asyncToGenerator3.default)(function* (obj, args, ctx, info) {
                /*  sanity check usage  */
                if (!ctx.scope) throw new Error("resume: only allowed on WebSocket connection");
                if (info && info.operation && info.operation.operation !== "mutation") throw new Error("resume: only allowed in GraphQL \"mutation\" operation");

                /*  determine parameters  */
                let cid = obj.cid;
                let sid = args.sid;

                /*  determine subscriptions  */
                yield _this5.keyval.acquire();
                let subscription = yield _this5.keyval.get(`sid:${sid},cid:${cid}`);
                if (subscription === undefined) {
                    yield _this5.keyval.release();
                    throw new Error(`no such subscription "${sid}"`);
                } else if (subscription !== "paused") {
                    yield _this5.keyval.release();
                    throw new Error(`subscription "${sid}" not paused`);
                }
                yield _this5.keyval.put(`sid:${sid},cid:${cid}`, "subscribed");
                yield _this5.keyval.release();

                /*  drop an event about the resuming  */
                ctx.scope.api.emit("scope-resume", { sid: sid, cid: cid });
                ctx.scope.api.emit("debug", `scope-resume sid=${sid} cid=${cid}`);

                /*  return no result  */
                return {};
            });

            return function (_x21, _x22, _x23, _x24) {
                return _ref6.apply(this, arguments);
            };
        })();
    }
}
exports.default = gtsSubscription;
